(function($, CRM, _, undefined) {

  var payment = {
    form: null,
    scriptName: 'CRM.payment',

    /**
     * Get the total amount on the form
     * @returns {number}
     */
    getTotalAmount: function() {
      var totalFee = 0.0;
      if (CRM.payment.isEventAdditionalParticipants()) {
        totalFee = null;
      }
      else if (document.getElementById('totalTaxAmount') !== null) {
        totalFee = parseFloat(this.calculateTaxAmount());
        this.debugging(this.scriptName, 'Calculated amount using internal calculateTaxAmount()');
      }
      else if (typeof calculateTotalFee == 'function') {
        // This is ONLY triggered in the following circumstances on a CiviCRM contribution page:
        // - With a priceset that allows a 0 amount to be selected.
        // - When Stripe is the ONLY payment processor configured on the page.
        totalFee = parseFloat(calculateTotalFee());
      }
      else if (this.getIsDrupalWebform()) {
        // This is how webform civicrm calculates the amount in webform_civicrm_payment.js
        $('.line-item:visible', '#wf-crm-billing-items').each(function() {
          totalFee += parseFloat($(this).data('amount'));
        });
      }
      else if (document.getElementById('total_amount')) {
        // The input#total_amount field exists on backend contribution forms
        totalFee = parseFloat(document.getElementById('total_amount').value);
      }
      this.debugging(this.scriptName, 'getTotalAmount: ' + totalFee);
      return totalFee;
    },

    /**
     * This is calculated in CRM/Contribute/Form/Contribution.tpl and is used to calculate the total
     *   amount with tax on backend submit contribution forms.
     *   The only way we can get the amount is by parsing the text field and extracting the final bit after the space.
     *   eg. "Amount including Tax: $ 4.50" gives us 4.50.
     *   The PHP side is responsible for converting money formats (we just parse to cents and remove any ,. chars).
     *
     * @returns {string|prototype.value|number}
     */
    calculateTaxAmount: function() {
      var totalTaxAmount = 0;
      if (document.getElementById('totalTaxAmount') === null) {
        return totalTaxAmount;
      }

      // If tax and invoicing is disabled totalTaxAmount div exists but is empty
      if (document.getElementById('totalTaxAmount').textContent.length === 0) {
        totalTaxAmount = document.getElementById('total_amount').value;
      }
      else {
        // Otherwise totalTaxAmount div contains a textual amount including currency symbol
        totalTaxAmount = document.getElementById('totalTaxAmount').textContent.split(' ').pop();
      }
      return totalTaxAmount;
    },

    /**
     * @returns {boolean}
     */
    isEventAdditionalParticipants: function() {
      if ((document.getElementById('additional_participants') !== null) &&
        (document.getElementById('additional_participants').value.length !== 0)) {
        this.debugging(this.scriptName, 'We don\'t know the final price - registering additional participants');
        return true;
      }
      return false;
    },

    getIsDrupalWebform: function() {
      // form class for drupal webform: webform-client-form (drupal 7); webform-submission-form (drupal 8)
      if (this.form !== null) {
        return this.form.classList.contains('webform-client-form') || this.form.classList.contains('webform-submission-form');
      }
      return false;
    },

    getBillingForm: function() {
      // If we have a stripe billing form on the page
      var billingFormID = $('div#card-element').closest('form').prop('id');
      if ((typeof billingFormID === 'undefined') || (!billingFormID.length)) {
        // If we have multiple payment processors to select and stripe is not currently loaded
        billingFormID = $('input[name=hidden_processor]').closest('form').prop('id');
      }
      // We have to use document.getElementById here so we have the right elementtype for appendChild()
      return document.getElementById(billingFormID);
    },

    /**
     * Output debug information
     * @param {string} scriptName
     * @param {string} errorCode
     */
    debugging: function(scriptName, errorCode) {
      // Uncomment the following to debug unexpected returns.
      if ((typeof(CRM.vars.stripe) === 'undefined') || (Boolean(CRM.vars.stripe.jsDebug) === true)) {
        console.log(new Date().toISOString() + ' ' + scriptName + ': ' + errorCode);
      }
    },

  };

  payment.form = payment.getBillingForm();
  if (typeof CRM.payment === 'undefined') {
    CRM.payment = payment;
  }
  else {
    $.extend(CRM.payment, payment);
  }

}(jQuery, CRM, _));
